import math
import typing as t
from argparse import Namespace, _SubParsersAction


class Parser:
    """Add the parsers to the subparser for the respective curve method.

    Attributes
    ----------
    _subparser : argparse._SubParsersAction
        The subparser, which contains all the parsers for the respective curve method (Must
        start with an underscore to skip the calling on initialization).

    """

    def __init__(self, subparser: _SubParsersAction):
        """Call all methods, which will add a parser for their respective computation method."""
        self._subparser = subparser
        for method_name in dir(self):
            if callable(getattr(self, method_name)) and not method_name.startswith("_"):
                getattr(self, method_name)()

    def sine(self) -> None:
        """Add parser for the sine wave."""
        parser = self._subparser.add_parser("sine", description="x(t) = t, y(t) = A*sin(2*pi*f*t)")
        parser.add_argument("-A", metavar="<A>", type=float, default=1, help="amplitude A (default: %(default)s)")
        parser.add_argument("-f", metavar="<f>", type=float, default=1, help="frequency f (default: %(default)s)")

    def lissajous(self) -> None:
        """Add parser for the lissajous curve."""
        parser = self._subparser.add_parser("lissajous", description="x(t, d) = sin(a*t + d), y(t) = sin(b*t)")
        parser.add_argument("-a", metavar="<a>", type=float, default=2, help="a in the ratio 'a/b' (default: %(default)s)")
        parser.add_argument("-b", metavar="<b>", type=float, default=3, help="b in the ratio 'a/b' (default: %(default)s)")

    def lissajous2(self) -> None:
        """Add parser for the second lissajous curve."""
        parser = self._subparser.add_parser("lissajous2", description="x(t, a) = sin(a*t + d), y(t) = sin(b*t), increase the ratio 'a/b' from 0 to n")
        parser.add_argument("-b", metavar="<b>", type=float, default=4, help="b in the ratio 'a/b', defines the number of waves (default: %(default)s)")
        parser.add_argument("-n", metavar="<n>", type=int, default=1, help="end for ratio 'a/b' (default: %(default)s)")


class F:
    """Calculate the values to display the animation.

    This class is needed to calculate the X- and Y-values of the animation in the
    terminal. When this class is initialized, all parsers are added to the subparser,
    where the curve to display can be chosen. Calling the resulting object will return
    the function to calculate the values for the curve.

    All these function must take one argument. This argument will move in the
    interval [-1, 1) to create the single frames.

    Attributes
    ----------
    parser : argparse._SubParsersAction
        The class containing the subparser, which contains all parsers for all curves.
    args : argparse.Namespace
        The namespace with all availables arguments (from both, the main parser
        and all parsers from the subparser).
    x : list[float]
        The list to draw the curve from x=-1 to x=1.

    Other attributes are specified in the method's respective computation
    or _const function.

    """

    def __init__(self, subparser: _SubParsersAction) -> None:
        """Add the parsers with the needed arguments for the respective curve."""
        self.parser = Parser(subparser)

    def __call__(self, args: Namespace) -> t.Callable:
        """Return the method to calculate the curve. Also calculate the needed constants."""
        self.args = args
        self.x = [x / self.args.resolution for x in range(-self.args.resolution, self.args.resolution+1)]
        if not args.curve:
            choices = []
            for choice in self.parser._subparser.choices:
                choices.append(choice)
            raise ValueError("Choose from {}".format(", ".join([f"'{c}'" for c in choices])))
        else:
            calc_const = getattr(self, "{}_{}".format(self.args.curve, "const"), None)
            if calc_const:
                calc_const()
            return getattr(self, self.args.curve)

    def sine(self, w: int) -> tuple[list[float], list[float]]:
        """Calculate a simple sine wave moving through the terminal."""
        y = [self.args.A * math.sin(self.args.f*(x+w) * math.pi) for x in self.x]
        return self.x, y

    def lissajous_const(self) -> None:
        """Calculate the constant for Y for the lissajous methods."""
        self.y = [math.sin(self.args.b * x * math.pi) for x in self.x]

    def lissajous(self, w: int) -> tuple[list[float], list[float]]:
        """Calculate X for the lissajous curves emulating slightly detuned frequencies."""
        x = [math.sin((self.args.a*x + w) * math.pi) for x in self.x]
        return x, self.y

    def lissajous2_const(self) -> None:
        """Call the method to calculate the contants for the lissajous2 method."""
        self.lissajous_const()

    def lissajous2(self, w: int) -> tuple[list[float], list[float]]:
        """Calculate X for the lissajous curves when increasing the ratio 'a/b'."""
        x = [math.sin(self.args.n/2*(w+1)*self.args.b * x * math.pi) for x in self.x]
        return x, self.y



import argparse
import time
from multiprocessing import Pipe, Process

from .cliplot import COLOR, CliPlot
from .f import F


def settings() -> None:
    """Print the settings for the rect-ratio and colors."""
    print()
    print("width/height")
    print("============")
    print()
    print("To display a square not like a rectangle (or a circle not like an ellipse),")
    print("count the pixels from")
    print("  - the first column to the last column. This is the width.")
    print("  - the first row the the last row. This is the height.")
    print("Calculate width/height. This is the value to use for '--rect-ratio'")
    print("to show the animation with proper proportions.")
    for _ in range(11):
        print("." * 10)

    print()
    print()
    print("colors")
    print("======")
    print()
    print("{:<15} | {:<5}".format("line-color", "background-color"))
    print("-" * 29)
    for key, color in COLOR.items():
        if key:
            print("\033[{}m{:<15}\033[0m | \033[{}m{}\033[0m".format(color, key, color+10, key))


def main() -> None:
    """Run the main function."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--settings", action="store_true", help="shows information about the rect-ratio and colors (default: %(default)s)")
    parser.add_argument("-p", "--period", metavar="<seconds>", type=float, default=3, help="time in seconds, until the figure finishes 1 cycle (f=1/T) (default: %(default)s)")
    parser.add_argument("-r", "--resolution", metavar="<res>", type=int, default=100, help="delta between two points (default: %(default)s)")
    parser.add_argument("-R", "--rect-ratio", metavar="<width/height>", type=float, default=0.5, help="ratio of width/height (depending on the font, needed that a circle appears like a circle and not like an ellipse) (default: %(default)s)")
    parser.add_argument("-s", "--symbol", metavar="<char>", type=str, default="#", help="symbol for the curve (default: '%(default)s')")
    parser.add_argument("-l", "--line-color", metavar="<color>", type=str, default=None, help="color of the curve (default: %(default)s)")
    parser.add_argument("-b", "--background-color", metavar="<color>", type=str, default=None, help="color of the background (default: %(default)s)")
    parser.add_argument("-g", "--grid-color", metavar="<color>", type=str, help="color of the grid (default: %(default)s)")
    parser.add_argument("-G", "--grid", action="store_true", help="enables grid (default: %(default)s)")
    parser.add_argument("-f", "--frame", metavar="<frame>", type=float, help="shows the curve at a certain frame without animation, value must be between -1 and 1 (default: %(default)s)")
    subparser = parser.add_subparsers(dest="curve")

    func = F(subparser)
    args = parser.parse_args()

    if args.settings:
        settings()
        return

    plt = CliPlot(args)
    f = func(args)

    if args.frame is not None:
        if args.grid or args.grid_color:
            plt.draw_grid()
        plt.scatter(*f(args.frame))
        plt.show()
        plt.colorize(reset=True)
        return

    p1, p2 = Pipe(False)
    if args.grid or args.grid_color:
        ani = plt.animate_with_grid
    else:
        ani = plt.animate
    proc = Process(target=ani, args=(p1,), daemon=True)
    proc.start()

    try:
        p, r = args.period, args.resolution
        w = -r
        while True:
            c_time = time.time_ns()
            p2.send(f(w/r))
            if w == r-1:
                w = -r
            else:
                w += 1
            while time.time_ns() <= c_time + (p/2/r*1e9):
                time.sleep(1e-6)

    except KeyboardInterrupt:
        proc.kill()
        while proc.is_alive():
            time.sleep(0.1)
        plt.colorize(reset=True)
        plt.paint_screen()


if __name__ == "__main__":
    main()



import os
import sys
import typing as t
from argparse import Namespace
from multiprocessing.connection import Connection

COLOR = {None: 0,
         "black": 30,
         "red": 31,
         "green": 32,
         "yellow": 33,
         "blue": 34,
         "magenta": 35,
         "cyan": 36,
         "white": 37,
         "bright-black": 90,
         "bright-red": 91,
         "bright-green": 92,
         "bright-yellow": 93,
         "bright-blue": 94,
         "bright-magenta": 95,
         "bright-cyan": 96,
         "bright-white": 97}


class CliPlot:
    """Print curves to the terminal.

    This class is to compute and print a matrix, that is a visual representation
    of scattered data points.
    The matrix looks like
    [0, 1, 2, 3, 4, ...]
    [0, 1, 2, 3, 4, ...]
    [0, 1, 2, 3, 4, ...]
    [0, 1, 2, 3, 4, ...]
    [...]

    but the data points are allowed in an interval from [-1, 1]. So these points
    are transformed in a way, that point P(0, 0) is in the middle of the matrix
    and not on the top left corner.

    Attributes
    ----------
    rect_ratio : float
        Ratio between width and height. Usually fonts have a smaller space between
        two chars in the same line (row) than two chars in the same column. With
        this setting, one can tune the coordinate system that a circle or square
        does not look distorted like an ellipse or rectangle.
    symbol : char
        The symbol, which is used to print the curves.
    line_color : str
        The foreground-color / the color of the line of the curve.
    background_color : str
        The color of the background of both the line and the coordinate system
        (plot area).
    grid_color : str
        The color of the grid.
    grid_symbol : char
        The symbol, which is used to print the grid.
    columns : int
        The number of the columns of the current terminal.
    rows : int
        The number of the rows -1 of the terminal. One is subtracted, because
        the lowest row should be left empty (otherwise the terminal starts
        scrolling). Well, testing it with lissajous2 and omitting '-1', it
        looks pretty cool in the terminal, when scrolling up again.
    w : int
        The max width, to have a scaled grid / coordinate system (based on
        rect_ratio).
    h : int
        The max height, to have a scaled grid / coordinate system (based on
        rect_ratio).
    wm : int
        The middle of the width of the scaled grid / cooridnate system.
        Used to draw the grid.
    hm : int
        The middle of the height of the scaled grid / cooridnate system.
        Used to draw the grid.
    matrix : list[list]
        This matrix contains the data, that will be plotted. More like, this
        matrix is a representation of the pixels in the terminal, where the
        'symbol' and grid is placed at its respective position. This matrix
        is then printed and the plot becomes visible in the terminal.

    """

    def __init__(self, args: Namespace) -> None:
        """Initialize the class."""
        self.rect_ratio = args.rect_ratio
        self.symbol = args.symbol
        self.line_color = args.line_color
        self.background_color = args.background_color
        self.grid_color = args.grid_color

        size = os.get_terminal_size()
        self.columns = size.columns
        self.rows = size.lines - 1
        self.scale()
        self.clear_matrix()
        self.colorize()

    def colorize(self, reset: bool = False) -> None:
        """Set the line- and background-color for the plot(s).

        This method sets the colors for the line (curve) and the background,
        as well as for the grid. It also resets the colors to the default
        terminal color, when 'True' is passed as argument.

        Parameters
        ----------
        reset, optional
            Reset the terminal's color to its defaults. The default is False.

        """
        if reset:
            sys.stdout.write("\033[0m")
            return

        c = []
        write = False
        if self.line_color:
            c.append(str(COLOR[self.line_color]))
            write = True
        if self.background_color:
            c.append(str(COLOR[self.background_color]+10))
            write = True
        if write:
            sys.stdout.write("\033[{}m".format(";".join(c)))

        if not self.grid_color:
            self.grid_symbol = "+"
        else:
            self.grid_symbol = "\033[{}m{}\033[{}m".format(COLOR[self.grid_color], "+", COLOR[self.line_color])

    def paint_screen(self) -> None:
        """Paint the screen by printing an empty matrix.

        This is useful at the end of the program to erase previously printed
        data points, so to bring the terminal back in a clear state.
        """
        matrix = [[" " for _ in range(self.columns)] for _ in range(int(self.rows))]
        self._move_cursor(1, 1)
        i = 1
        for line in matrix:
            self._move_cursor(1, i)
            sys.stdout.write("".join(line) + "\n")
            i += 1
        sys.stdout.flush()

    def scale(self) -> None:
        """Scale the matrix based on 'rect_ratio'.

        When printing a matrix where all pixels are filled with data points, a
        square will appear rather than a rectangle. This is to ensure that the
        curves are correctly displayed with 1:1 proportions.

        """
        if (x_max := int(self.rows / self.rect_ratio)) <= self.columns:
            self.w = x_max
            self.h = self.rows
        else:
            self.w = self.columns
            self.h = int(self.columns * self.rect_ratio)
        self.wm = self.x(0)
        self.hm = self.y(0)

    def clear_matrix(self) -> None:
        """Remove all data points from the matrix."""
        self.matrix = [[" " for _ in range(self.w)] for _ in range(int(self.h))]

    def draw_grid(self) -> None:
        """Draw a grid in the middle of the plot."""
        for i in range(self.w):
            self.matrix[self.hm][i] = self.grid_symbol
        for i in range(self.h):
            self.matrix[i][self.wm] = self.grid_symbol

    @staticmethod
    def _move_cursor(x: int, y: int) -> None:
        """Move the terminal cursor to the specified position.

        This is needed to create an animation, because the lines get
        overwritten for each new frame. This is more efficient than clearing
        the terminal for every frame of the animation.

        Parameters
        ----------
        x : int
            Row in the terminal.
        y : int
            Column in the terminal.

        """
        sys.stdout.write(f"\033[{y};{x}H")
        sys.stdout.flush()

    def x(self, x: float) -> int:
        """Transform a value in [-1, 1] to their respective column."""
        return round((self.w-1) / 2 * (x + 1))

    def y(self, y: float) -> int:
        """Transform a value in [-1, 1] to their respective row."""
        return round((self.h-1) / 2 * (1 - y))

    def scatter(self, x: t.Union[list, float], y: t.Union[list, float]) -> None:
        """Write a symbol in the matrix at position (x, y)."""
        if type(x) is list and type(y) is list and (leng := len(x)) == len(y):
            for i in range(leng):
                if abs(y[i]) > 1:
                    continue
                self.matrix[self.y(y[i])][self.x(x[i])] = self.symbol
        elif type(x) is not list and type(y) is not list:
            self.matrix[self.y(y)][self.x(x)] = self.symbol  # type: ignore[arg-type]

        elif type(x) is list and type(y) is list and len(x) != len(y):
            raise Exception("Lengths must be the same.")
        else:
            raise Exception("Datatypes must be the same. Both list or int.")

    def show(self) -> None:
        """Print the matrix to the terminal, starting in the top left corner."""
        i = 1
        for line in self.matrix:
            self._move_cursor(1, i)
            sys.stdout.write("".join(line) + "\n")
            i += 1
        sys.stdout.flush()

    def animate(self, p: Connection) -> None:
        """Print the matrix with the newly received data points to the terminal."""
        while True:
            recv = p.recv()  # type: tuple[t.Union[list, int], t.Union[list, int]]
            self.clear_matrix()
            self.scatter(*recv)
            self.show()

    def animate_with_grid(self, p: Connection) -> None:
        """Behave as animate, but also print the grid to the terminal."""
        while True:
            recv = p.recv()  # type: tuple[t.Union[list, int], t.Union[list, int]]
            self.clear_matrix()
            self.draw_grid()
            self.scatter(*recv)
            self.show()
